﻿using System;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapter 1
    
    public class MMath
    {
        #region Permutations
        /// <summary>
        /// Returns the number of possible permutations of k objects from a set of n object. The order of the chosen objects does matter.
        /// </summary>
        /// <param name="n">Number of objects</param>
        /// <param name="k">Number of objects chosen</param>
        /// <returns></returns>
        public static double Permutations(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            //long permut = 1;
            //for (int i = n; i > (n - k); i--)
            //    permut = checked(permut * i);\
            double logPermut = LogPermutations(n, k);
            return Math.Exp(logPermut);
        }

        internal static double LogPermutations(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            double permut = 0;
            for (int i = n; i > (n - k); i--)
                permut += Math.Log(i);
            return permut;
        }
        #endregion

        #region Combinations
        /// <summary>
        /// Returns the number of possible combinations of k objects from a set of n object. The order of the chosen objects does not matter.
        /// </summary>
        /// <param name="n">Number of objects</param>
        /// <param name="k">Number of objects chosen</param>
        public static double Combinations(int n, int k)
        {
            double logCombin = LogCombin(n, k);
            double combin = Math.Exp(logCombin);
            return Math.Round(combin, 0);
        }

        internal static double LogCombin(int n, int k)
        {
            return LogFactorial(n) - LogFactorial(k) - LogFactorial(n - k);
        }

        private static double LogFactorial(int x)
        {
            double lf = 0;
            for (int i = 2; i <= x; i++)
                lf += Math.Log(i);
            return lf;
        }

        //This method can handle values up to n = 70, but is very slow
        private static long CombinationsOld(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            if (k > n) throw new System.ArgumentException("Combin not defined for k greater than n. Cannot choose more objects than there are objects.");

            if (k == 0 || n == k) return 1;
            return checked(CombinationsOld(n - 1, k - 1) + CombinationsOld(n - 1, k));
        }

        //This method is faster, but cannot handle n > 20
        private static long CombinationsFast(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            if (k > n - k) return CombinationsFast(n, n - k);

            long c = 1;
            for (int i = n; i > (n - k); i--)
                c = checked(c * i);
            long kFact = Factorial(k);
            return c / kFact;
        }
        #endregion


        /// <summary>
        /// Returns n! 
        /// 0! = 1,otherwise n! = n * (n-1) * (n-2) * ... * 2 * 1,
        /// </summary>
        public static long Factorial(int n)
        {
            if (n < 0) throw new System.ArgumentException("Factorial not defined for negative n");
            if (n > 20) throw new System.ArgumentException("Answer will exceed max long");
            long fact = 1;
            for (int i = n; i > 0; i--)
                fact *= i;
            return fact;
        }

        /// <summary>
        /// Integral from 0 to infinity of e^(-t) * t^(z-1) dt
        /// </summary>
        public static double GammaFunction(double z)
        {
            //based on approximation from NIST Handbook of Mathematical Function. 2010. Cambridge University Press.
            
            if(z > 143)
                throw new System.ArgumentException("Cannot currently compute gamma function for z > 143");

            //need z<21 or z<11 becaus Factorial(x) will not work for x>20
            if (z > 0 && z < 21 && z % 1.0 == 0)    
                return Factorial((int)Math.Round(z - 1, 0));
            if (z > 0 && z < 11 && z % 0.5 == 0)
            {
                int n = (int)z;
                return Math.Sqrt(Math.PI) * Factorial(2 * n) / (Math.Pow(4, n) * Factorial(n));
            }

            double sqrtTwoPi = 2.5066282746310002;
            double d = 1.0 + 1.0 / (12 * z) + 1.0 / (288 * z * z) - 139.0 / (51840 * Math.Pow(z, 3)) - 571.0 / (2488320 * Math.Pow(z, 4))
                + 163879.0 / (209018880 * Math.Pow(z, 5)) + 5246819.0 / (75246796800 * Math.Pow(z, 6));
            double g = Math.Pow(z, z - 0.5) * Math.Exp(-z) * sqrtTwoPi * d;
            return g;
        }

        /// <summary>
        /// B(a, b) = Integral(0,1) of (t^(a-1))*((1-t)^(b-1)) dt. Also know as Euler integral. 
        /// </summary>
        public static double BetaFunction(double a, double b)
        {
            if (a + b > 143)
            {
                if (b > 20)
                    return 2.5066282746310002 * Math.Pow(a, a - 0.5) * Math.Pow(b, b - 0.5) / Math.Pow(a + b, a + b - 0.5);
                else
                    return GammaFunction(b) * Math.Pow(a, -b);
            }
                
            return GammaFunction(a) * GammaFunction(b) / GammaFunction(a + b);
        }

        /// <summary>
        /// B(x, a, b) = Integral(0,x) of (t^(a-1))*((1-t)^(b-1)) dt. Also know as Euler integral. 
        /// </summary>
        public static double IncompleteBetaFunction(double x, double a, double b)
        {
            if (x == 0) return 0;
            if (x == 1) return BetaFunction(a, b);

            //The numerical integration is less accurate for high values of x, so variable step sizes are used with 
            //progressively more steps for higher values of x.
            if (x <= 0.9)
            {
                return PartialIncompleteBetaFunction(0, x, a, b);
            }
            else if (x <= 0.99)
            {
                double s1 = PartialIncompleteBetaFunction(0, 0.9, a, b);
                double s2 = PartialIncompleteBetaFunction(0.90, x, a, b);
                return s1 + s2;
            }
            else
            {
                double s1 = PartialIncompleteBetaFunction(0, 0.9, a, b);
                double s2 = PartialIncompleteBetaFunction(0.9, 0.99, a, b);
                double s3 = PartialIncompleteBetaFunction(0.99, x, a, b);
                return s1 + s2 + s3;
            }
        }

        private static double PartialIncompleteBetaFunction(double xL, double xU, double a, double b)
        {
            if (xU == xL) return 0.0;
            if (xU < xL) return double.NaN;

            int nSteps = 1000;
            if (a < 1) nSteps = 80000;

            double dt = (xU - xL) / nSteps;
            double t = xL + 0.5 * dt;
            double sum = 0;
            for (int i = 0; i < nSteps; i++)
            {
                sum += Math.Pow(t, a - 1) * Math.Pow(1.0 - t, b - 1) * dt;
                t += dt;
            }
            return sum;
        }

        public static double RegularizedIncompleteBetaFunction(double x, double a, double b)
        {
            if (x == 0) return 0.0;
            if (x == 1) return 1.0;

            if (a % 1.0 == 0 && b % 1.0 == 0 && (a + b) < 22)
                return RegularizedIncompleteBetaFunction(x, (int)a, (int)b);

            //if (x > 0.5)
            //    return 1.0 - RegularizedIncompleteBetaFunction(1.0 - x, b, a);    //thought this would be faster, but it slows down Students t CDF, because b = 0.5. Swithcing b to a, slows PartialIncompleteBetaFunction, which adds extra steps for a < 1. Does not seem to impact accuracy.
            return IncompleteBetaFunction(x, a, b) / BetaFunction(a, b);
        }

        private static double RegularizedIncompleteBetaFunction(double x, int a, int b)
        {
            double s = 0;
            for(int i=a; i<a+b; i++)
            {
                s += Math.Pow(x, i) * Math.Pow(1.0 - x, a + b - 1 - i) / (Factorial(i) * Factorial(a + b - 1 - i));
            }
            s *= Factorial(a+b-1);
            return s;
        }

        /// <summary>
        /// Returns x(x+1)(x+2)...(x+n-1) for n>0; for n=0 returns 1.
        /// </summary>
        public static double PochhammerFunction(double x, double n)
        {
            if (n < 0) return double.NaN;
            if (n == 0) return 1;
            double p = 1;
            for (int i = 0; i < n; i++)
                p *= (x + i);
            return p;
        }
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.