﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapters 6 and 7

    public class Matrix
    {
        public readonly int nRows;              //the number of rows in the matrix
        public readonly int nColumns;           //the number of columns in the matrix
        private readonly double[,] elements;    //the elements of the matrix

        #region Constructors
        /// <summary>
        /// Creates a new matrix with elements corresponding to the elements of the array.
        /// </summary>
        /// <param name="array">The elements of this array will be the initial elements of the matrix.</param>
        public Matrix(double[,] array)
        {
            nRows = array.GetLength(0);
            nColumns = array.GetLength(1);
            if(nRows == 0 || nColumns == 0)
                throw new System.ArgumentException("The elements of the array must have at least one row and one column");
            this.elements = (double[,])array.Clone();
        }

        /// <summary>
        /// Creates a vector, a matrix with one column.
        /// </summary>
        /// <param name="array">The elements of this array will be the initial elements of the matrix.</param>
        public Matrix(double[] array)
        {
            nRows = array.Length;
            if (nRows == 0)
                throw new System.ArgumentException("The array must have at least one element");
            nColumns = 1;
            this.elements = new double[nRows, 1];
            for (int r = 0; r < nRows; r++)
                this.elements[r, 0] = array[r];
        }

        /// <summary>
        /// Creates a new matrix, nRows x nColumns, with all elements equal to zero.
        /// </summary>
        /// <param name="nRows">Number of rows.</param>
        /// <param name="nColumns">Number of columns.</param>
        public Matrix(int nRows, int nColumns)
        {
            this.nRows = nRows;
            this.nColumns = nColumns;
            elements = new double[nRows, nColumns];
        }
        #endregion

        /// <summary>
        /// Creates an nxn identity matrix.
        /// </summary>
        /// <param name="n">Size of idenity matrix = number of rows = number of columns.</param>
        public static Matrix IdentityMatrix(int n)
        {
            Matrix M = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                M[i, i] = 1.0;
            return M;
        }


        /// <summary>
        /// Returns element [r,c] from the matrix.
        /// </summary>
        /// <param name="r">row index</param>
        /// <param name="c">column index</param>
        public double this[int r, int c]
        {
            get { return this.elements[r, c]; }
            set { this.elements[r, c] = value; }
        }

        public Matrix Copy()
        {
            Matrix M = new Matrix(this.nRows, this.nColumns);
            for (int r = 0; r < this.nRows; r++)
                for (int c = 0; c < this.nColumns; c++)
                    M[r, c] = this[r, c];
            return M;
        }

        /// <summary>
        /// Returns the sum of matrix M and the existing matrix.
        /// </summary>
        /// <param name="M">Must be the same size as the existing matrix.</param>
        public Matrix Add(Matrix M)
        {
            if (this.nRows != M.nRows || this.nColumns != M.nColumns)
                throw new System.ArgumentException("Cannot add. The matrices must be the same size.");
            Matrix sum = new Matrix(nRows, nColumns);
            for (int r = 0; r < nRows; r++)
                for (int c = 0; c < nColumns; c++)
                    sum[r, c] = this[r, c] + M[r, c];
            return sum;
        }

        /// <summary>
        /// Matrix subtraction. X - M.
        /// </summary>
        /// <param name="M">The matrix to be subtracted.</param>
        public Matrix Subtract(Matrix M)
        {
            Matrix minusM = M.Multiply(-1);
            return this.Add(minusM);
        }

        /// <summary>
        /// Returns the transpose of the existing matrix.
        /// </summary>
        public Matrix Transpose()
        {
            Matrix T = new Matrix(nColumns, nRows);
            for (int r = 0; r < nRows; r++)
                for (int c = 0; c < nColumns; c++)
                    T[c, r] = this[r, c];
            return T;
        }

        /// <summary>
        /// Matrix multiplication. Post multiplies the existing matrix by another matrix, M.
        /// </summary>
        /// <param name="M">The number of rows in m should be equal to the number of columns in the existing matrix.</param>
        public Matrix Multiply(Matrix M)
        {
            if (this.nColumns != M.nRows)
                throw new System.ArgumentException("Cannot multiply. Number of columns in first matrix must equal the number of rows in the second.");
            int innerD = this.nColumns;
            Matrix product = new Matrix(this.nRows, M.nColumns);
            for (int r = 0; r < product.nRows; r++)
                for (int c = 0; c < product.nColumns; c++)
                    for (int i = 0; i < innerD; i++)
                        product[r, c] += this[r, i] * M[i, c];
            return product;
        }

        /// <summary>
        /// Scalar multiplication. Multiplies the existing matrix by d.
        /// </summary>
        /// <param name="d">The scalar by which the matrix is being multiplied.</param>
        /// <returns></returns>
        public Matrix Multiply(double d)
        {
            Matrix product = new Matrix(this.nRows, this.nColumns);
            for (int r = 0; r < product.nRows; r++)
                for (int c = 0; c < product.nColumns; c++)
                    product[r, c] = d * this[r, c];
            return product;
        }

        #region Matrix Decomposition
        /// <summary>
        /// Decomposes a symmetric square matrix into upper and lower triangular matrices, L and U, where L = U'.
        /// </summary>
        /// <param name="U">Upper triangular matrix</param>
        /// <param name="L">Lower triangular matrix</param>
        public void CholeskyDecomposition(out Matrix L, out Matrix U)
        {
            if (this.nRows != this.nColumns)
                throw new System.ArgumentException("Can only decompose square matrix. Number of rows and columns must be equal.");
            if (!this.IsSymmetric())
                throw new System.ArgumentException("Can only use Cholesky with a symmetric matrix.");
            int n = this.nRows;
            L = new Matrix(n, n);

            for(int r = 0; r < n; r++)
                for (int c = 0; c <= r; c++)
                {
                    if (r == c)
                    {
                        double d = this[r, c];
                        for (int i = 0; i < r; i++)
                            d -= L[r, i] * L[r, i];
                        L[r, c] = Math.Sqrt(d);
                    }
                    else
                    {
                        double d = this[r, c];
                        for (int i = 0; i < c; i++)
                            d -= L[r, i] * L[c, i];
                        d /= L[c, c];
                        L[r, c] = d;
                    }
                }

            U = L.Transpose();
        }

        /// <summary>
        /// Returns true if matrix is square and X[r, c] = X[c, r] for all r and c.
        /// </summary>
        /// <returns></returns>
        public bool IsSymmetric()
        {
            if (nRows != nColumns) return false;
            for (int r = 0; r < nRows; r++)
                for (int c = 0; c < r; c++)
                    if (this[r, c] != this[c, r]) return false;
            return true;
        }

        /// <summary>
        /// Decomposes a square matrix into upper and lower triangular matrices, L and U, such that X = LU.
        /// </summary>
        /// <param name="U">Upper triangular matrix</param>
        /// <param name="L">Lower triangular matrix</param>
        public void LuDecomposition(out Matrix L, out Matrix U)
        {
            if (this.nRows != this.nColumns)
                throw new System.ArgumentException("Can only decompose square matrix. Number of rows and columns must be equal.");

            int n = this.nRows;
            L = new Matrix(n, n);
            U = this.Copy();

            for (int r = 0; r < n; r++)
            {
                L[r, r] = U[r, r];
                for (int c = 0; c < n; c++)
                    U[r, c] = U[r, c] / L[r, r];
                
                for (int i = r + 1; i < n; i++)
                {
                    L[i, r] = U[i, r];
                    for (int j = 0; j < n; j++)
                        U[i, j] -= L[i, r] * U[r, j];
                }
            }
        }
        #endregion

        #region Matrix inversion
        /// <summary>
        /// Calculates the inverse of the existing matrix. Matrix must be square.
        /// </summary>
        public Matrix Inverse()
        {
            //This method first decomposes the existing matrix into a lower and upper tringular matrix. These
            //can be more easily inverted.
            //There may be more precise or efficient algorithms, but this is very easy to follow.

            if (this.nRows != this.nColumns)
                throw new System.ArgumentException("Can only invert square matrix. Number of rows and columns must be equal.");

            Matrix L, U;
            this.LuDecomposition(out L, out U);

            Matrix Linv = L.LowerTriangularInverse();
            Matrix Uinv = U.Transpose().LowerTriangularInverse().Transpose();

            Matrix inverse = Uinv.Multiply(Linv);

            return inverse;
        }

        public bool isCloseToIdentity(int precision)
        {
            if (nRows != nColumns) return false;
            double d = Math.Pow(10, -precision);
            for(int r = 0; r<nRows; r++)
                for (int c = 0; c < nColumns; c++)
                {
                    if (r == c)
                    {
                        if (Math.Abs(this[r, c] - 1) > d) return false;
                    }
                    else
                    {
                        if (Math.Abs(this[r, c]) > d) return false;
                    }
                }
            return true;
        }

        public bool IsLowerTriangular()
        {
            return this.Transpose().IsUpperTriangular();
        }

        public bool IsUpperTriangular()
        {
            if (this.nRows != this.nColumns)
                throw new System.ArgumentException("Upper and lower triangular only defined for square matrices. Number of rows and columns must be equal.");
            int n = this.nRows;
            for (int r = 1; r < n; r++)
            {
                for (int c = 0; c < r; c++)
                    if (this[r, c] != 0)
                        return false;
            }
            return true;
        }

        /// <summary>
        /// Calculates the inverse of a lower triangular matrix. Used to calculate general inverses.
        /// </summary>
        /// <returns></returns>
        private Matrix LowerTriangularInverse()
        {
            if (!this.IsLowerTriangular())
                throw new System.Exception("Can only use with lower triangular matrix");
            int n = this.nRows;

            Matrix D = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                D[i, i] = 1.0 / this[i, i];
            Matrix C = D.Multiply(this);

            //needed to prevent occasional rounding error
            for (int i = 0; i < n; i++)
                C[i, i] = 1.0; 
            
            Matrix Cinv = C.LowerTriangularInverseOneDiagonal();
            Matrix M = Cinv.Multiply(D);

            return M;
        }

        /// <summary>
        /// Used to calculate LowerTriangularInverse, which is used to calculate inverses.
        /// </summary>
        private Matrix LowerTriangularInverseOneDiagonal()
        {
            //The inverse of a lower triangular matrix with ones on the diagonal is also
            //lower triangular with ones on the diagonal.
            //This problem is relatively easy to solve relative and can be used to solve the more
            //general inversion problem.

            if (!this.IsLowerTriangular())
                throw new System.Exception("Can only use with lower triangular matrix");
            int n = this.nRows;
            for (int i = 0; i < n; i++)
                if (this[i, i] != 1)
                    throw new System.Exception("Can only use where all diagonal elements are 1");

            Matrix M = new Matrix(n, n);
            for (int i = 0; i < n; i++)
                M[i, i] = 1;

            for(int d = 1; d < n; d++)
                for (int r = d; r < n; r++)
                {
                    int c = r - d;
                    for (int i = 0; i < r; i++)
                        M[r, c] -= this[r, i] * M[i, c];
                }

            return M;
        }
        #endregion

    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.