﻿using System;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapters 2, 3 and 10

    public class Moments
    {
        #region Mean
        /// <summary>
        /// Returns the mean of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the mean.</param>
        public static double Mean(double[] array)
        {
            return Mean(array, 1.0);
        }

        /// <summary>
        /// Returns the mean of an array.
        /// </summary>
        /// <param name="array">Array of data fro which we are calculating the mean.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double Mean(double[] array, double decayFactor)
        {
            int n = array.Length;
            double m = 0.0;
            double d = 1.0;
            for (int i = 0; i < n; i++)
            {
                m += d * array[n - 1 - i];
                d *= decayFactor;
            }

            m *= GeometricSeries.InverseSumOfGeometricSeries(decayFactor, n);
            return m;
        }

        /// <summary>
        /// Returns the mean of an array.
        /// </summary>
        /// <param name="array">Array of data fro which we are calculating the mean.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public static double Mean(double[] array, double decayFactor, int length)
        {
            double[] subArray = Tools.MostRecentValues(array, length);
            return Mean(subArray, decayFactor);
        }
        #endregion

        #region StandardDeviation
        /// <summary>
        /// Returns the sample standard deviation of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the standard deviation.</param>
        public static double StandardDeviation(double[] array)
        {
            return StandardDeviation(array, 1.0);
        }

        /// <summary>
        /// Returns the sample variance of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the variance.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double StandardDeviation(double[] array, double decayFactor)
        {
            double v = Variance(array, decayFactor);
            return Math.Sqrt(v);
        }

        /// <summary>
        /// Returns the sample variance of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the variance.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public static double StandardDeviation(double[] array, double decayFactor, int length)
        {
            double[] subArray = Tools.MostRecentValues(array, length);
            return StandardDeviation(subArray, decayFactor);
        }
        #endregion

        #region Incremental Standard Deviation
        /// <summary>
        /// Returns the incremental standard deviation of a variable. Incremental standard deviation is to standard deviation what incremental VaR is to VaR.
        /// </summary>
        /// <param name="portfolioArray">Return array of portfolio. The last element (index = n-1), is the most recent.</param>
        /// <param name="subPortfolioArray">Return array of the sub-portfolio for which incremental standard deviation is being measured. The last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1. Weigth on the last element in arrays is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double IncrementalStandardDeviation(double[] portfolioArray, double[] subPortfolioArray, double decayFactor)
        {
            if (portfolioArray.Length != subPortfolioArray.Length)
                throw new System.ArgumentException("Arrays must be the same length");

            int len = portfolioArray.Length;
            double[] otherArray = new double[len];
            for (int i = 0; i < len; i++)
                otherArray[i] = portfolioArray[i] - subPortfolioArray[i];

            double varianceSubPort = Variance(subPortfolioArray, decayFactor);
            double covarSubOther = Covariance(subPortfolioArray, otherArray, decayFactor);
            double sdPort = StandardDeviation(portfolioArray, decayFactor);
            double iStandardDeviation = (varianceSubPort + covarSubOther) / sdPort;
            return iStandardDeviation;
        }

        /// <summary>
        /// Returns the incremental standard deviation of a variable. Incremental standard deviation is to standard deviation what incremental VaR is to VaR.
        /// </summary>
        /// <param name="portfolioArray">Return array of portfolio. The last element (index = n-1), is the most recent.</param>
        /// <param name="subPortfolioArray">Return array of the sub-portfolio for which incremental standard deviation is being measured. The last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1. Weigth on the last element in arrays is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public static double IncrementalStandardDeviation(double[] portfolioArray, double[] subPortfolioArray, double decayFactor, int length)
        {
            double[] portfolioArrayLen = Tools.MostRecentValues(portfolioArray, length);
            double[] subPortfolioArrayLen = Tools.MostRecentValues(subPortfolioArray, length);
            return IncrementalStandardDeviation(portfolioArrayLen, subPortfolioArrayLen, length);
        }
        #endregion

        #region Variance
        /// <summary>
        /// Returns the sample variance of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the variance.</param>
        public static double Variance(double[] array)
        {
            return Variance(array, 1.0);
        }

        /// <summary>
        /// Returns the sample variance of an array.
        /// </summary>
        /// <param name="array">Array of data fro which we are calculating the variance.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double Variance(double[] array, double decayFactor)
        {
            int n = array.Length;
            double v = 0.0;
            double d = 1.0;
            for (int i = 0; i < n; i++)
            {
                v += d * array[n - 1 - i] * array[n - 1 - i];
                d *= decayFactor;
            }

            double m = Mean(array, decayFactor);
            double S1 = GeometricSeries.SumOfGeometricSeries(decayFactor, n);
            double S2 = GeometricSeries.SumOfGeometricSeries(decayFactor * decayFactor, n);
            double A = S1 / (S1 * S1 - S2);
            double B = S1 * A;
            v = A * v - B * m * m;
            return v;
        }

        /// <summary>
        /// Returns the sample variance of an array.
        /// </summary>
        /// <param name="array">Array of data fro which we are calculating the variance.  For time series, the last element (index = n-1), is the most recent.</param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in array is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        /// <returns></returns>
        public static double Variance(double[] array, double decayFactor, int length)
        {
            double[] subArray = Tools.MostRecentValues(array, length);
            return Variance(subArray, decayFactor);
        }
        #endregion

        #region Covariance and Correlation
        /// <summary>
        /// Returns the sample covariance between two arrays.
        /// Arrays should be of equal length, and contain more than one element.
        /// </summary>
        /// <param name="array1"></param>
        /// <param name="array2"></param>
        public static double Covariance(double[] array1, double[] array2)
        {
            if (array1.Length != array2.Length)
                throw new System.ArgumentException("Arrays must be the same length");

            int n = array1.Length;
            if (n == 1) return double.NaN;
            double m1 = Mean(array1);
            double m2 = Mean(array2);
            
            
            double covar = 0.0;
            for (int i = 0; i < n; i++)
                covar += (array1[i] - m1) * (array2[i] - m2);
            covar /= (n - 1);
            return covar;
        }

        /// <summary>
        /// Returns the sample covariance between two arrays.
        /// Arrays should be of equal length, and contain more than one element.
        /// </summary>
        /// <param name="array1"></param>
        /// <param name="array2"></param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in arrays is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double Covariance(double[] array1, double[] array2, double decayFactor)
        {
            if (array1.Length != array2.Length)
                throw new System.ArgumentException("Arrays must be the same length");

            int n = array1.Length;
            if (n == 1) return double.NaN;
            double covar = 0.0;
            double d = 1.0;
            for (int i = 0; i < n; i++)
            {
                covar += d * array1[n - 1 - i] * array2[n - 1 - i];
                d *= decayFactor;
            }

            double m1 = Mean(array1, decayFactor);
            double m2 = Mean(array2, decayFactor);
            double S1 = GeometricSeries.SumOfGeometricSeries(decayFactor, n);
            double S2 = GeometricSeries.SumOfGeometricSeries(decayFactor * decayFactor, n);
            double A = S1 / (S1 * S1 - S2);
            double B = S1 * A;
            covar = A * covar - B * m1 * m2;
            return covar;
        }

        /// <summary>
        /// Returns the correlation between two arrays.
        /// Arrays should be of equal length, and contain more than one element.
        /// </summary>
        /// <param name="array1"></param>
        /// <param name="array2"></param>
        public static double Correlation(double[] array1, double[] array2)
        {
            double covar = Covariance(array1, array2);
            double sd1 = StandardDeviation(array1);
            double sd2 = StandardDeviation(array2);
            return covar / (sd1 * sd2);
        }

        /// <summary>
        /// Returns the correlation between two arrays.
        /// Arrays should be of equal length, and contain more than one element.
        /// </summary>
        /// <param name="array1"></param>
        /// <param name="array2"></param>
        /// <param name="decayFactor">In most applications, the decay factor is between 0 and 1.  Weigth on the last element in arrays is 1.0, the 2nd to last element d, 3rd to last d^2, ...</param>
        public static double Correlation(double[] array1, double[] array2, double decayFactor)
        {
            double covar = Covariance(array1, array2, decayFactor);
            double sd1 = StandardDeviation(array1, decayFactor);
            double sd2 = StandardDeviation(array2, decayFactor);
            return covar / (sd1 * sd2);
        }
        #endregion

        #region Skewness
        /// <summary>
        /// Returns the sample skewness of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the skewness.</param>
        public static double Skewness(double[] array)
        {
            int n = array.Length;
            if(n < 3) return double.NaN;
            double m = Mean(array);
            double sd = StandardDeviation(array);

            double skew = 0.0;
            for (int i = 0; i < n; i++)
                skew += Math.Pow(array[i] - m, 3.0);
            skew /= Math.Pow(sd, 3.0);
            skew = n * skew / ((n - 1) * (n - 2));
            return skew;
        }
        #endregion

        #region Kurtosis
        /// <summary>
        /// Returns the sample kurtosis of an array.
        /// Note: This is not excess kurtosis.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the kurtosis.</param>
        public static double Kurtosis(double[] array)
        {
            int n = array.Length;
            if (n < 4) return double.NaN;
            double m = Mean(array);
            double variance = Variance(array);

            double kurt = 0.0;
            for (int i = 0; i < n; i++)
                kurt += Math.Pow(array[i] - m, 4.0);
            kurt /= (variance * variance);
            kurt *= (n * (n + 1.0)) / ((n - 1.0) * (n - 2.0) * (n - 3.0));
            return kurt;
        }

        /// <summary>
        /// Returns the sample excess kurtosis of an array.
        /// </summary>
        /// <param name="array">Array of data for which we are calculating the excess kurtosis.</param>
        public static double ExcessKurtosis(double[] array)
        {
            double kurt = Kurtosis(array);
            int n = array.Length;
            kurt -= 3 * (n - 1.0) * (n - 1.0) / ((n - 2.0) * (n - 3.0));
            return kurt;
        }
        #endregion
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.