﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapters 8 and 10

    public class Regression
    {
        #region Public Assessors
        public Matrix Y { get; private set; }
        public Matrix X { get; private set; }

        /// <summary>
        /// Array of parameter estimates.
        /// </summary>
        public ParameterEstimate[] Betas { get; private set; }

        /// <summary>
        /// Total sum of squares.
        /// </summary>
        public double TSS { get; private set; }

        /// <summary>
        /// Explained sum of squares.
        /// </summary>
        public double ESS { get { return TSS - RSS; } }

        /// <summary>
        /// Residual sum of squares.
        /// </summary>
        public double RSS { get; private set; }

        /// <summary>
        /// R-squared is also known as the coefficient of determination.
        /// </summary>
        public double rSquared { get; private set; }

        /// <summary>
        /// Adjusted R-squared.
        /// </summary>
        public double AdjustedRSquared { get; private set; }

        /// <summary>
        /// F-statistic.
        /// </summary>
        public double fStatistic { get; private set; }

        /// <summary>
        /// Probability that the F-statistic is not significant. Values closer to 0% indicate that the regression parameters are more statistically significant.
        /// </summary>
        public double fValue { get; private set; }
        #endregion

        #region Constuctors
        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variable, or regressor.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        public Regression(double[] yData, double[] xData, bool includeConstant)
        {
            double[,] xDataTwoD = new double[xData.Length, 1];
            for (int i = 0; i < xData.Length; i++)
                xDataTwoD[i, 0] = xData[i];

            PopulateMatrices(yData, xDataTwoD, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand. Vector is Tx1, where T is the number of time periods.</param>
        /// <param name="xData">Data for the independent variables, or regressors. Matrix is TxN, where T is the number of time periods, and N is the number of independent variables.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        public Regression(double[] yData, double[,] xData, bool includeConstant)
        {
            PopulateMatrices(yData, xData, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variable, or regressor.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public Regression(double[] yData, double[] xData, bool includeConstant, int length)
        {
            double[] yDataSubSet = Tools.MostRecentValues(yData, length);
            double[] xDataSubSet = Tools.MostRecentValues(xData, length);

            double[,] xDataTwoDSubSet = new double[length, 1];
            for (int i = 0; i < length; i++)
                xDataTwoDSubSet[i, 0] = xDataSubSet[i];

            PopulateMatrices(yDataSubSet, xDataTwoDSubSet, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand. Vector is Tx1, where T is the number of time periods.</param>
        /// <param name="xData">Data for the independent variables, or regressors. Matrix is TxN, where T is the number of time periods, and N is the number of independent variables.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public Regression(double[] yData, double[,] xData, bool includeConstant, int length)
        {
            double[] yDataSubSet = Tools.MostRecentValues(yData, length);
            double[,] xDataSubSet = Tools.MostRecentValues(xData, length);

            PopulateMatrices(yDataSubSet, xDataSubSet, includeConstant);
            CalculateStatistic();
        }
        #endregion

        private void PopulateMatrices(double[] yData, double[,] xData, bool includeConstant)
        {
            Y = new Matrix(yData);
            if (includeConstant)
            {
                double[,] xDataPlus = new double[xData.GetLength(0), xData.GetLength(1) + 1];
                for (int r = 0; r < xDataPlus.GetLength(0); r++)
                {
                    xDataPlus[r, 0] = 1.0;
                    for (int c = 1; c < xDataPlus.GetLength(1); c++)
                        xDataPlus[r, c] = xData[r, c - 1];
                }
                X = new Matrix(xDataPlus);
            }
            else
            {
                X = new Matrix(xData);
            }
            if (Y.nRows != X.nRows)
                throw new System.ArgumentException("Dependent and independent data do not have the same number of rows.");

        }

        private void CalculateStatistic()
        {
            Matrix XprimeXinv = X.Transpose().Multiply(X).Inverse();
            Matrix betaValues = XprimeXinv.Multiply(X.Transpose()).Multiply(Y);

            int t = Y.nRows;
            int n = betaValues.nRows;

            TSS = Y.Transpose().Multiply(Y)[0, 0];
            Matrix expectedValueY = X.Multiply(betaValues);
            Matrix errors = expectedValueY.Subtract(Y);
            RSS = errors.Transpose().Multiply(errors)[0, 0];
            rSquared = 1 - RSS / TSS;
            AdjustedRSquared = 1 - (1 - rSquared) * (t - 1) / (t - n);

            
            double varianceOfErrorTerm = errors.Transpose().Multiply(errors)[0, 0] / (t - n);
            Betas = new ParameterEstimate[n];
            for (int i = 0; i < n; i++)
            {
                double sigma = Math.Sqrt(varianceOfErrorTerm * XprimeXinv[i, i]);
                double tStat = Math.Abs(betaValues[i, 0] - 0) / sigma;
                double pValue = 1.0 - Distributions.StudentsTCumulativeDensityFunction(tStat, t - n);
                Betas[i] = new ParameterEstimate(betaValues[i, 0], sigma, tStat, pValue);
            }

            fStatistic = (rSquared / (n - 1)) / ((1 - rSquared) / (t - n));
            fValue = 1.0 - Distributions.FCumulativeDensityFunction(fStatistic, n - 1, t - n);
        }
    }

    public class WeightedLeastSquares
    {
        #region Private fields
        private double _decayFactor;
        private Matrix _Xstar;
        private Matrix _Ystar;
        #endregion

        #region Pubulic Assessors
        public Matrix Y { get; private set; }
        public Matrix X { get; private set; }

        /// <summary>
        /// Array of parameter estimates.
        /// </summary>
        public ParameterEstimate[] Betas { get; private set; }

        /// <summary>
        /// Total sum of squares.
        /// </summary>
        public double TSS { get; private set; }

        /// <summary>
        /// Explained sum of squares.
        /// </summary>
        public double ESS { get { return TSS - RSS; } }

        /// <summary>
        /// Residual sum of squares.
        /// </summary>
        public double RSS { get; private set; }

        /// <summary>
        /// R-squared is also known as the coefficient of determination.
        /// </summary>
        public double rSquared { get; private set; }

        /// <summary>
        /// Adjusted R-squared.
        /// </summary>
        public double AdjustedRSquared { get; private set; }

        /// <summary>
        /// F-statistic.
        /// </summary>
        public double fStatistic { get; private set; }

        /// <summary>
        /// Probability that the F-statistic is not significant. Values closer to 0% indicate that the regression parameters are more statistically significant.
        /// </summary>
        public double fValue { get; private set; }
        #endregion

        #region Constuctors
        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variable, or regressor.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="decayFactor">More recent data (higher index value) is weighted more heavily. Most recent data point has weight d, then d^2, then d^3, ... . If decay factor is 1.0 then weighted least squares is equivalent to ordinary least squares.</param>
        public WeightedLeastSquares(double[] yData, double[] xData, bool includeConstant, double decayFactor)
        {
            _decayFactor = decayFactor;

            double[,] xDataTwoD = new double[xData.Length, 1];
            for (int i = 0; i < xData.Length; i++)
                xDataTwoD[i, 0] = xData[i];

            PopulateMatrices(yData, xDataTwoD, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// <param name="yData">Data for the dependent variable, or regressand. Vector is Tx1, where T is the number of time periods.</param>
        /// <param name="xData">Data for the independent variables, or regressors. Matrix is TxN, where T is the number of time periods, and N is the number of independent variables.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="decayFactor">More recent data (higher index value) is weighted more heavily. Most recent data point has weight d, then d^2, then d^3, ... . If decay factor is 1.0 then weighted least squares is equivalent to ordinary least squares.</param>
        public WeightedLeastSquares(double[] yData, double[,] xData, bool includeConstant, double decayFactor)
        {
            _decayFactor = decayFactor;

            PopulateMatrices(yData, xData, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variable, or regressor.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="decayFactor">More recent data (higher index value) is weighted more heavily. Most recent data point has weight d, then d^2, then d^3, ... . If decay factor is 1.0 then weighted least squares is equivalent to ordinary least squares.</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public WeightedLeastSquares(double[] yData, double[] xData, bool includeConstant, double decayFactor, int length)
        {
            _decayFactor = decayFactor;

            double[] yDataSubSet = Tools.MostRecentValues(yData, length);
            double[] xDataSubSet = Tools.MostRecentValues(xData, length);

            double[,] xDataTwoDSubSet = new double[length, 1];
            for (int i = 0; i < length; i++)
                xDataTwoDSubSet[i, 0] = xDataSubSet[i];

            PopulateMatrices(yDataSubSet, xDataTwoDSubSet, includeConstant);
            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand. Vector is Tx1, where T is the number of time periods.</param>
        /// <param name="xData">Data for the independent variables, or regressors. Matrix is TxN, where T is the number of time periods, and N is the number of independent variables.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        /// <param name="decayFactor">More recent data (higher index value) is weighted more heavily. Most recent data point has weight d, then d^2, then d^3, ... . If decay factor is 1.0 then weighted least squares is equivalent to ordinary least squares.</param>
        /// <param name="length">Window length. Method uses the most recent n points, n = length.</param>
        public WeightedLeastSquares(double[] yData, double[,] xData, bool includeConstant, double decayFactor, int length)
        {
            _decayFactor = decayFactor;

            double[] yDataSubSet = Tools.MostRecentValues(yData, length);
            double[,] xDataSubSet = Tools.MostRecentValues(xData, length);

            PopulateMatrices(yDataSubSet, xDataSubSet, includeConstant);
            CalculateStatistic();
        }
        #endregion

        private void PopulateMatrices(double[] yData, double[,] xData, bool includeConstant)
        {
            Y = new Matrix(yData);
            if (includeConstant)
            {
                double[,] xDataPlus = new double[xData.GetLength(0), xData.GetLength(1) + 1];
                for (int r = 0; r < xDataPlus.GetLength(0); r++)
                {
                    xDataPlus[r, 0] = 1.0;
                    for (int c = 1; c < xDataPlus.GetLength(1); c++)
                        xDataPlus[r, c] = xData[r, c - 1];
                }
                X = new Matrix(xDataPlus);
            }
            else
            {
                X = new Matrix(xData);
            }
            if (Y.nRows != X.nRows)
                throw new System.ArgumentException("Dependent and independent data do not have the same number of rows.");

            int t = Y.nRows;
            Matrix W = new Matrix(t, t);
            double d = 1.0;
            for (int i = t - 1; i >= 0; i--)
            {
                for (int j = 0; j < t; j++)
                {
                    if (i == j)
                        W[i, j] = d;
                    else
                        W[i, j] = 0;
                }
                d *= _decayFactor;
            }

            _Xstar = W.Multiply(X);
            _Ystar = W.Multiply(Y);
        }

        private void CalculateStatistic()
        {
            Matrix XprimeXinv = _Xstar.Transpose().Multiply(_Xstar).Inverse();
            Matrix betaValues = XprimeXinv.Multiply(_Xstar.Transpose()).Multiply(_Ystar);

            int t = Y.nRows;
            int n = betaValues.nRows;

            TSS = _Ystar.Transpose().Multiply(_Ystar)[0, 0];
            Matrix expectedValueY = _Xstar.Multiply(betaValues);
            Matrix errors = expectedValueY.Subtract(_Ystar);
            RSS = errors.Transpose().Multiply(errors)[0, 0];
            rSquared = 1 - RSS / TSS;
            AdjustedRSquared = 1 - (1 - rSquared) * (t - 1) / (t - n);

            double varianceOfErrorTerm = errors.Transpose().Multiply(errors)[0, 0] / (t - n);
            Betas = new ParameterEstimate[n];
            for (int i = 0; i < n; i++)
            {
                double sigma = Math.Sqrt(varianceOfErrorTerm * XprimeXinv[i, i]);
                double tStat = Math.Abs(betaValues[i, 0] - 0) / sigma;
                double pValue = 1.0 - Distributions.StudentsTCumulativeDensityFunction(tStat, t - n);
                Betas[i] = new ParameterEstimate(betaValues[i, 0], sigma, tStat, pValue);
            }

            fStatistic = (rSquared / (n - 1)) / ((1 - rSquared) / (t - n));
            fValue = 1.0 - Distributions.FCumulativeDensityFunction(fStatistic, n - 1, t - n);
        }
    }

    public class ParameterEstimate
    {
        public double Value { get; private set; }
        public double StandardDeviation { get; private set; }

        /// <summary>
        /// This is an indication of how statistically significant the parameter value is. It is tested against the null hypothesis that the parameter is zero. The value is specified to always be positive. |[beta]|/[standard deviation of beta].
        /// </summary>
        public double tStatistic { get; private set; }
        
        /// <summary>
        /// The probability of observing a parameter value this different than zero by chance. Always less than 50%. Closer to 0% is more statistically significant.
        /// </summary>
        public double pValue { get; private set; }

        internal ParameterEstimate(double value, double standardDeviation, double tStatistic, double pValue)
        {
            Value = value;
            StandardDeviation = standardDeviation;
            this.tStatistic = tStatistic;
            this.pValue = pValue;
        }
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.